import re


FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')

def to_url(name):
    sub = FIRST_CAP_RE.sub(r'\1-\2', name)
    return ALL_CAP_RE.sub(r'\1-\2', sub).lower()


class Logger(object):
    DEBUG = 1
    MINIMAL = 0

    level = 1

    @staticmethod
    def debug(msg, *objects):
        if Logger.level >= Logger.DEBUG:
            print msg % objects



from db import QueryObject, MemoryQueryObject, UnQLiteQueryObject
from utils import *
from error import Error404
import flask
from types import GeneratorType


class Resource(object):
    def __init__(self, app, name=None):
        if name is not None:
            self.__class__.__name__ = name

        self.app = app
        self.objects = UnQLiteQueryObject(self.get_name())

    def on_create(self):
        pass

    def register(self, endpoint, method, route=None):
        if route is None:
            route = endpoint.__name__

        route = self.app.route_factory.clean_relative_route(route)
        route = self.get_name() + '/' + route
        route = self.app.route_factory.create_route(route)

        def wrapper():
            if method == 'GET':
                params = flask.request.args
            else:
                params = flask.request.get_json()

            response = endpoint(params)

            if isinstance(response, GeneratorType):
                response = list(response)

            return self.app._response_middleware(response)

        self.app._add_route(route, wrapper, endpoint.__name__, method)

    def list(self):
        """
        GET <resource>/ - returns a list of resources.
        """
        return self.objects.all()

    def post(self, body):
        """
        POST <resource>/ - create a new resource and returns it.
        """
        return self.objects.create(body)

    def get(self, pk):
        """
        GET <resource>/:pk/ - returns a single resource.
        """
        obj = self.objects.get(pk)
        if obj is None:
            raise Error404()
        return obj

    def delete(self, pk):
        """
        DELETE <resource>/:pk/ - delete a resource.
        """
        return self.objects.delete(pk)

    def put(self, pk, body):
        """
        PUT <resource>/:pk/ - updates a resource.
        """
        return self.objects.update(pk, body)

    def get_name(self):
        return to_url(self.__class__.__name__)



from unqlite import UnQLite


class QueryObject(object):
    def __init__(self, collection_name):
        self.collection_name = collection_name

    def all(self):
        pass

    def create(self, body):
        pass

    def get(self, pk):
        pass

    def delete(self, pk):
        pass

    def update(self, pk, body):
        pass


class MemoryQueryObject(QueryObject):
    def __init__(self, collection_name):
        super(MemoryQueryObject, self).__init__(collection_name)

        self.objects = dict()
        self.current_pk = 0

    def all(self):
        return self.objects.values()

    def create(self, body):
        if 'pk' in body:
            # throw exception.
            pass

        body = self._assign_pk(body)
        self.objects[self.current_pk] = body
        self.current_pk += 1
        return body

    def get(self, pk):
        if pk in self.objects:
            return self.objects[pk]
        return None

    def delete(self, pk):
        obj = self.get(pk)
        del self.objects[pk]
        return obj

    def update(self, pk, body):
        self.objects[pk].update(body)
        return self.objects[pk]

    def _assign_pk(self, body):
        body = body.copy()
        body['pk'] = self.current_pk
        return body


class UnQLiteQueryObject(QueryObject):
    db = UnQLite('db.unqlite')

    def __init__(self, collection_name):
        super(UnQLiteQueryObject, self).__init__(collection_name)

        self.collection = self.db.collection(self.collection_name)

        # create if the collection does not exists.
        self.collection.create()

    def all(self):
        return self.collection.all()

    def create(self, body, pk=None):
        """
        :param: pk pass a pk to create or update if it exists.
        """
        if pk is not None:
            previous = self.collection[pk]
            if previous is not None:
                return self.update(pk, body)

        pk = self.collection.store(body)
        return self.get(pk)

    def get(self, pk):
        return self.collection[pk]

    def delete(self, pk):
        obj = self.get(pk)
        self.collection.delete(pk)
        return obj

    def update(self, pk, body):
        self.collection.update(pk, body)
        return self.get(pk)

    def filter(self, condition):
        return self.collection.filter(condition)



class RouteFactory(object):
    def __init__(self, base_route):
        self.base_route = self._clean_base(base_route)

        self.list_format = "%s/"
        self.post_format = "%s/"
        self.get_format = "%s/<pk>/"
        self.delete_format = "%s/<pk>/"
        self.put_format = "%s/<pk>/"

    def create_list(self, instance):
        return self.create_route(self.list_format % instance.get_name())

    def create_post(self, instance):
        return self.create_route(self.post_format % instance.get_name())

    def create_get(self, instance):
        return self.create_route(self.get_format % instance.get_name())

    def create_delete(self, instance):
        return self.create_route(self.delete_format % instance.get_name())

    def create_put(self, instance):
        return self.create_route(self.put_format % instance.get_name())

    def create_route(self, relative):
        return '/' + self.base_route + '/' + relative

    def _clean_base(self, base):
        if base.startswith('/'):
            base = base[1:]

        if base.endswith('/'):
            base = base[:-1]

        return base

    def clean_relative_route(self, route):
        if route.startswith('/'):
            route = route[1:]

        if not route.endswith('/'):
            route += '/'

        return route



#!/usr/bin/env python
from setuptools import setup

setup(name='ezrpy',
      version='0.1.3',
      description='Automatic REST API for Prototypes',
      author='Marcelo Jara',
      author_email='marjara35@gmail.com',
      keywords=['rest', 'prototype', 'api'],
      url='https://github.com/mijara/ezrpy',
      packages=['ezrpy'],
      install_requires=[
          'flask',
      ],
      )



import flask
import re
import json
import inspect
import flask_cors

from handle import Resource
from route import RouteFactory
from utils import Logger
from error import Error404
from db import UnQLiteQueryObject


class App(object):
    def __init__(self, base_route):
        self.base_route = base_route
        self.debug = True

        self.app = flask.Flask(__name__)
        self.route_factory = RouteFactory(base_route)

        flask_cors.CORS(self.app)

        @self.app.after_request
        def after_request(response):
            response.headers.add('Access-Control-Allow-Origin', '*')
            response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
            response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE')
            return response

    def add_resource(self, resource):
        if inspect.isclass(resource) and issubclass(resource, Resource):
            instance = resource(self)
            instance.on_create()
        elif isinstance(resource, str):
            instance = Resource(self, name=resource)

        self._inspect_list(instance)
        self._inspect_post(instance)
        self._inspect_get(instance)
        self._inspect_delete(instance)
        self._inspect_put(instance)

        return instance

    def _add_route(self, route, endpoint, as_name, method):
        if as_name is not None:
            endpoint.__name__ = as_name
        self.app.route(route, methods=[method])(endpoint)
        Logger.debug("Create route for %s: %s", endpoint.__name__, route)

    def _inspect_list(self, instance):
        route = self.route_factory.create_list(instance)
        name = instance.get_name()

        def endpoint():
            return self._response_middleware(instance.list())

        self._add_route(route, endpoint, '%s_list' % name, 'GET')

    def _inspect_post(self, instance):
        route = self.route_factory.create_post(instance)
        name = instance.get_name()

        def endpoint():
            body = flask.request.get_json()
            return self._response_middleware(instance.post(body))

        self._add_route(route, endpoint, '%s_post' % name, 'POST')

    def _inspect_get(self, instance):
        route = self.route_factory.create_get(instance)
        name = instance.get_name()

        def endpoint(pk):
            try:
                return self._response_middleware(instance.get(long(pk)))
            except Error404:
                return self._response_error(404, 'Object not found')

        self._add_route(route, endpoint, '%s_get' % name, 'GET')

    def _inspect_delete(self, instance):
        route = self.route_factory.create_delete(instance)
        name = instance.get_name()

        def endpoint(pk):
            return self._response_middleware(instance.delete(long(pk)))

        self._add_route(route, endpoint, '%s_delete' % name, 'DELETE')

    def _inspect_put(self, instance):
        route = self.route_factory.create_delete(instance)
        name = instance.get_name()

        def endpoint(pk):
            body = flask.request.get_json()
            return self._response_middleware(instance.put(long(pk), body))

        self._add_route(route, endpoint, '%s_put' % name, 'PUT')

    def _response_middleware(self, response):
        return flask.Response(json.dumps(response), mimetype='application/json')

    def _response_error(self, status, message):
        return flask.Response(json.dumps({
            'message': message
        }), mimetype='application/json', status=404)

    def run(self, host='127.0.0.1', port=5000):
        self.app.run(host=host, port=port, threaded=True)

        UnQLiteQueryObject.db.close()
